#pragma once
#ifndef SEGWT_H
#define SEGWT_H
#include <cstdlib>
#include <cstring>
#include <cstdarg>
#include <cstdio>

/* zpsob jak alokovat pam zarovnanou na 16byte boundary, ukazatel pro uvolnn se uchov "ped" alokovanou pamt
POZOR! Pokud malloc u zarovnv na 32byte, tenhle pstup to me zhorit.

void *mem = malloc(1024+15+sizeof(void*));
if (!mem) return mem;
void *ptr = ((char*)mem+sizeof(void*)+15) & ~ (size_t)0x0F;
((void**)ptr)[-1] = mem;
return ptr;

if (ptr) free(((void**)ptr)[-1]);
*/


#define pow2(x) ((1)<<(x))  // makro power of two by bit shift (for interers)
#define modPow2(x,pow2var) ((x)&(pow2var-1)) // power two modulo by bit masking
#define isPowOf2(x) ((x)&(x-1))==0 // power of two test by bit opeation
#define min(X,Y) ((X) < (Y) ? (X) : (Y))
#define max(X,Y) ((X) > (Y) ? (X) : (Y))
#define MAX_PARAMS 21

// interface definition method for coefficient processing
class IWaveletCoeffProcessor{
public:
	// coefficient processing method
	virtual void process(float** in, float** out,int* coefLens, int J)=0;
	virtual void setParams(float * params, int paramLen) = 0;
	virtual void getParams(float* params, int paramLen) = 0;
};

// interface adding access to input samples e.g. for noise variance estimation
class IWaveletCoeffInputProcessor:public IWaveletCoeffProcessor{
public:
	virtual void process(float** in, float** out,int* coefLens, int J,float* input, int inputLen)=0;
};


// DefaultProcessor simply copies input coefficients to output
class DefaultProcessor: public IWaveletCoeffProcessor{
public:
	void process(float** in, float** out,int* coefLens,int J);
	void setParams(float * params, int paramLen) {}
	void getParams(float* params, int paramLen) {}
};

// FilterProcessor multiplies each subband j with filterArray[j] value
class FilterProcessor: public IWaveletCoeffProcessor{
private:
	float filterArray[MAX_PARAMS];
public:
	FilterProcessor(){	for(int j=0;j<MAX_PARAMS;j++)	filterArray[j] = 1.0f;	}
	void process(float** in, float** out,int* coefLens,int J);
	void setParams(float * params, int paramLen) {
		memcpy(filterArray,params,paramLen*sizeof(float));
	}
	void getParams(float* params, int paramLen) {
		memcpy(params,filterArray,paramLen*sizeof(float));
	}

	void setFilterValue(int j,float value){
		filterArray[j] = value;
	}
	void setFilterValues(int Jplus1,float* value){
		memcpy(filterArray,value,Jplus1*sizeof(float));
	}
};

// ThresholdProcessor thresholds (hard) each subband j with thresholdArray[j] value
class ThresholdProcessor: public IWaveletCoeffProcessor{
private:
	float thresholdArray[MAX_PARAMS];
	void hardThreshold(float* in,float* out,int Len,float threshold);
public:
	ThresholdProcessor(){ for(int j=0;j<MAX_PARAMS;j++)	thresholdArray[j] = 0.0f;	}

	void process(float** in, float** out,int* coefLens,int J);
	void setParams(float * params, int paramLen) {
		memcpy(thresholdArray,params,paramLen*sizeof(float));
	}
	void getParams(float* params, int paramLen) {
		memcpy(params,thresholdArray,paramLen*sizeof(float));
	}
	void setThresholdValue(int j,float value){
		thresholdArray[j] = value;
	}
	void setThresholdValues(int Jplus1,float value){
		for(int j=0;j<Jplus1;j++)
		 thresholdArray[j] = value;
	}
	void setThresholdValues(int Jplus1,float* value){
		memcpy(thresholdArray,value,Jplus1*sizeof(float));
	}
	
};

// RandomProcessor should multiply random deatil subband with value mult
class RandomProcessor: public IWaveletCoeffProcessor{
private:
	float multArray[MAX_PARAMS];
public:
	RandomProcessor(){{ for(int j=0;j<MAX_PARAMS;j++)	multArray[j] = 1.0f;	}}
	void process(float** in, float** out,int* coefLens,int J);
	
	void setParams(float * params, int paramLen) {
		memcpy(multArray,params,paramLen*sizeof(float));
	}
	void getParams(float* params, int paramLen) {
		memcpy(params,multArray,paramLen*sizeof(float));
	}

	void setMultValue(int j_, float mult_){
		multArray[j_] = mult_;
	}

	void setThresholdValues(int Jplus1,float value){
		for(int j=0;j<Jplus1;j++)
		 multArray[j] = value;
	}

	void setThresholdValues(int Jplus1,float* values){
		memcpy(multArray,values,Jplus1*sizeof(float));
	}

};


/**
Class representing wavelet filters. All data types are float
*/
class FloatWfilter{

public:
	/**
	Enum defining wavelet filters. Names adopted from Matlab Wavelet Toolbox
	*/
	enum Type {
        haar=0, db1, db2, db3, db4, db5, db6, db7, db8, db9, db10, sym2, sym3, sym4,
        sym5, sym6, sym7, sym8, coif1, coif2, coif3, coif4, coif5, bior11, bior13,
        bior15, bior22, bior24, bior26, bior28, bior31, bior33, bior35, bior37, bior39,
        bior44, bior55, bior68, cust, WAVELET_NUMBER };
	
	/**
	Static array of strings holding string names for members of enum Type
	*/
	const static  char* typeNames[];
	/**
	Default constructor makes haar filters.
	*/
	FloatWfilter();
	/**
	Constructor makes "name" wavelet filters.
	*/
	FloatWfilter(Type name);
	/**
	Constructor makes custom wavelet filters. Passed arrays are copied to internal memory.
	*/
	FloatWfilter(float* lo_a, float* hi_a, float* lo_r,float* hi_r,int filtLen);

	~FloatWfilter();
	FloatWfilter(const FloatWfilter& old);
	const FloatWfilter& operator=(const FloatWfilter& rhs);

	// analysis low-pass filter
	float* lo_a;
	// analysis high-pass filter
	float* hi_a;
	// synthetis low-pass filter
	float* lo_r;
	// synthetis high.pass filter
	float* hi_r;
	// filters length
	int filtLen;
	// name of wavelet filters
	Type name;

private: 
	void initialize(Type name);
	float* toArray(int n_args , ...);
};


class FloatSegDWT{
private:
	// depth of decomposition
	int J; 
	// wavelet filters
	FloatWfilter* wavelet;

	unsigned long Sn;  
	unsigned long segLen;

	// buffers holding intermediate results
	float* buffer; 
	int bufferLen;
	float* buffer2;

	// maximal and minimal left extensions
	int maxLeftExt;
	int minLeftExt;

	// buffers holding intermediate results
	float* tempA;
	int tempALen;int tempALenMax;
	float* tempD;
	int tempDLen;int tempDLenMax;
	
	float* tempIn;
	int tempInLen; int tempInLenMax;
	float* tempIn2;
	int tempIn2Len; int tempInLen2Max;

	// for convinience pointers to memory to be freed
	float** subbandsToFree;

	// array holds overlap afrer reconstruction
	int overlapLen;
	float* overlap;

	// array holds previous input samples
	int prevLen;
	float* prev;

	// template pattern - walelet coefficient processing is injected in procesor->process method
	IWaveletCoeffProcessor* procesor;

	// arrays of coefficients
	float** coefOut;
	// array of modified coefficients
	float** coefOutMod;
	int* coefOutLen;
	int* coefOutModLen;

	// flag  if external wavelet coefficient processor was provided (wont be freed)
	bool externalProcessor;

	
public:
	/***************************************************
	******************Implementations*******************
	***************************************************/

	// returns filter length
	int getFiltLen(){ return wavelet->filtLen;	}
	
	void setWaveletProcessor(IWaveletCoeffProcessor* procesor_){
		if(procesor && !externalProcessor) delete procesor;
		procesor = procesor_;
		externalProcessor = true;
	}

	IWaveletCoeffProcessor* getWaveletProcessor(){
		return procesor;
	}

	/***************************************************
	****************  Prototyps  ***********************
	***************************************************/

	FloatSegDWT(FloatWfilter::Type waveletType,int newJ=1);

	FloatSegDWT(FloatWfilter* newWavelet,int newJ=1);

	~FloatSegDWT();

	void process(float* in, float* out, int inLen, unsigned long Sn=0);

	void flush();

	void setJ(int newJ);

	void setWavelet(FloatWfilter* newWavelet);

	void forwardOLA(float* in, int inLen,
					float* out[],int outLen[],
					float* outOverlap[],int outOverlapLen[]);

	void forwardOLS(float* in,int inLen,
					float* out[],int outLen[],
					unsigned long Sn=0);

	void inverseOLA(float* in[], int inLen[],
								float* out, int outLen,
								unsigned long Sn=0);


private:
	// prototyps
	void convSub(float* in, int inLen,
							 float* outLo, float* outHi, int outLen,
							 float* filtLo, float* filtHi, int filtLen,
							 int skip = 0);

	void upConv(float* inLo, float* inHi, int inLoLen,
				float* out, int outLen,
				float* filtLo, float* filtHi, int filtLen,
				int skip = 0);

	void initialize(int newJ);

		// there is definetely no memory allocated at adresses out[j]
	// for convinience, there is little bit more memory allocated (and set to zero)
public:	void allocateSubbands(float** out,int* outLengths,unsigned long Sn,int s,int J);
private:
	// there is possibly memory allocated at out[j]
	// 1) difference: additional r(J-k) samples from the beginning
	// 2) 
	void reAllocateOutSubbands(float** out,int* outLengths,unsigned long Sn,int s,int J);

	void reAllocateInSubbands(float** out,int* outLengths,unsigned long Sn,int s,int J);

	/***************************************************
	******************Implementations*******************
	***************************************************/



	int ncoefs(unsigned long Sn, int s, int j){
		unsigned long Snplus1 = Sn + (unsigned long) s;
		unsigned long pow2j = (unsigned long) pow2(j);
		return (int) (Snplus1/pow2j - Sn/pow2j);
	}

	int nextPow2(int in){
		in = (in >> 1) | in;
		in = (in >> 2) | in;
		in = (in >> 4) | in;
		in = (in >> 8) | in;
		in = (in >> 16) | in;
		in++;
		return in;
	}

	int dwtLen(int s, int m){
        return (int)((s+m-1)/2.0);
    }

	int leftExtNoRight(unsigned long Sn,int m,int J){
		if(Sn==0) return 0;
		else
		return (int)( (unsigned long)((pow2(J)-1)*(m-2)) + modPow2(Sn,(unsigned long)pow2(J)));
	}

	int maxLeftExtNoRight(int m,int J){
		return (int)((pow2(J)-1)*(m-1));
	}
	int minLeftExtNoRight(int m,int J){
		return (int)((pow2(J)-1)*(m-2));
	}

	int discardLeftExtNoRight(unsigned long Sn,int m,int J,int j){
		if(J==j || Sn==0) return 0;
		return (int)( (pow2(J-j)-1)*(m-2)+ ((int)modPow2(Sn,(unsigned long)pow2(J)))/pow2(j) );
	}

	int appendLeftExtNoRight(unsigned long Sn,int m,int J,int j){
		if(J==j) return 0;
		return (int)( (pow2(J-j)-1)*(m-2)+ ((int)modPow2(Sn,(unsigned long)pow2(J)))/pow2(j) );
	}

	int addMaxLeftZerosInv(int J, int j, int m){
		return (pow2(J-j)-1)*(m-1);
	}

	int addLeftZerosInv(unsigned long Sn, int J, int j, int m){
		return (pow2(J-j)-1)*(m-2) + (int)(modPow2(Sn,(unsigned long)(pow2(J)))/((unsigned long)pow2(j)));
	}


};







#endif 